/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

size_t output_tail_count;
size_t output_head_count;
char *output_file = "-";

#define PID_CMD 0
#define PID_ERR 1
#define PID_OUT 2
#define _PID_MAX 3

struct procent
{
	const char *descr;
	pid_t pid;
	int status;
};

struct procent proctab[_PID_MAX] = {
	{ "output command", -1, 0 },
	{ "stderr redirector", -1, 0 },
	{ "stdout redirector", -1, 0 }
};

char *output_format = "%i\n";
char *begin_output_message;
char *end_output_message;

FILE *outfile;
int refcnt;
char **refnames;

void
slb_add_outfd(fd_set *fdset)
{
	if (outfile)
		FD_SET(fileno(outfile), fdset);
}

void
slb_set_output_defaults()
{
	output_file = "-";
}

/* %i id
   %h hostname
   %[0][w][.p]w weight
   %[0][w][.p]{var} var
*/
static void
format_srv(struct slb_server *srv)
{
	char *fmt = output_format;
	char *fbuf = xmalloc(strlen(output_format) + 2);
	char *p;
	
	while (fmt) {
		size_t len = strcspn(fmt, "%\\");
		if (len > 0) {
			fwrite(fmt, len, 1, outfile);
			fmt += len;
		}
		if (!*fmt)
			break;
		if (*fmt == '%') {
			int fn = 0;
			
			fbuf[fn++] = *fmt++;

			if (*fmt == '%') {
				fputc('%', outfile);
				fmt++;
				continue;
			}
			
			if (*fmt == '-' || *fmt == '+')
				fbuf[fn++] = *fmt++;
			if (*fmt == '0')
				fbuf[fn++] = *fmt++;
			while (c_isdigit(*fmt))
				fbuf[fn++] = *fmt++;
			if (*fmt == '.') {
				fbuf[fn++] = *fmt++;
				while (c_isdigit(*fmt))
					fbuf[fn++] = *fmt++;
			}
			
			switch (*fmt++) {
			case 'i':
				fbuf[fn++] = 's';
				fbuf[fn] = 0;
				fprintf(outfile, fbuf, srv->id);
				break;

			case 'h':
				fbuf[fn++] = 's';
				fbuf[fn] = 0;
				fprintf(outfile, fbuf, srv->peer);
				break;

			case 'w':
				fbuf[fn++] = 'f';
				fbuf[fn] = 0;
				fprintf(outfile, fbuf, srv->weight);
				break;

			case '(':
				p = fmt;
				for (; *fmt && *fmt != ')'; fmt++)
					;
				if (*fmt != ')') {
					fbuf[fn] = 0;
					fprintf(outfile, "%s", fbuf);
					fmt = p;
				} else {
					size_t n = fmt - p;
					struct slb_macro *mac;

					mac = macro_lookup(&srv->macros,
							   p, n, 0);
					if (!mac) {
						logmsg(LOG_WARNING,
						       "undefined macro in "
						       "format: %.*s",
						       (int) n, p);
						fbuf[fn] = 0;
						fprintf(outfile, "%s(", fbuf);
						fmt = p;
					} else {
						fbuf[fn++] = 's';
						fbuf[fn] = 0;
						fprintf(outfile, fbuf,
							mac->mac_exp);
						fmt++;
					}
				}
				break;
				
			case '{':
				p = fmt;
				for (; *fmt && *fmt != '}'; fmt++)
					;
				if (*fmt != '}') {
					fbuf[fn] = 0;
					fprintf(outfile, "%s", fbuf);
					fmt = p;
				} else {
					size_t n = fmt - p;
					struct slb_varinstance *vi;

					vi = varinst_lookup(&srv->varinst,
							    p, n, 0);
					if (!vi) {
						logmsg(LOG_WARNING,
						       "unknown variable in "
						       "format: %.*s",
						       (int) n, p);
						fbuf[fn] = 0;
						fprintf(outfile, "%s{", fbuf);
						fmt = p;
					} else {
						fbuf[fn++] = 'f';
						fbuf[fn] = 0;
						fprintf(outfile, fbuf,
							vi->vi_value);
						fmt++;
					}
				}
				break;

			default:
				fbuf[fn] = 0;
				fprintf(outfile, "%s%c", fbuf, fmt[-1]);
				logmsg(LOG_WARNING,
				       "unknown format specification in "
				       "format: %s%c",
				       fbuf, fmt[-1]);
				break;
			}
		}
	}
	free(fbuf);	
}

void
collect_output_refs()
{
	int i;
	char *fmt = output_format;
	struct slb_varinst_tab vit;

	memset(&vit, 0, sizeof(vit));
	
	while (fmt) {
		size_t len = strcspn(fmt, "%\\");
		if (len > 0)
			fmt += len;
		if (!*fmt)
			break;
		if (*fmt == '%') {
			fmt++;

			if (*fmt == '%') {
				fmt++;
				continue;
			}
			
			if (*fmt == '-' || *fmt == '+')
				fmt++;
			if (*fmt == '0')
				fmt++;
			while (c_isdigit(*fmt))
				fmt++;
			if (*fmt == '.') {
				fmt++;
				while (c_isdigit(*fmt))
					fmt++;
			}
			
			if (*fmt++ == '{') {
				char *p = fmt;
				for (; *fmt && *fmt != '}'; fmt++)
					;
				if (*fmt != '}') {
					fmt = p;
				} else {
					size_t n = fmt - p;
					struct slb_varinstance *vi;
					vi = varinst_lookup(&vit, p, n, 1);
					if (vi && !vi->vi_name) {
						char *q = xmalloc(n + 1);
						memcpy(q, p, n);
						q[n] = 0;
						vi->vi_name = q;
					}
					fmt++;
				}
			}
		}
	}

	refcnt = vit.vit_max;
	refnames = xcalloc(refcnt + 1, sizeof(refnames[0]));
	for (i = 0; i < refcnt; i++)
		refnames[i] = (char*) vit.vit_tab[i].vi_name;
	refnames[i] = NULL;
}

int
is_output_refname(const char *name)
{
	int i;
	
	for (i = 0; i < refcnt; i++) {
		if (strcmp(refnames[i], name) == 0)
			return 1;
	}
	return 0;
}

static size_t
format_output(size_t i, size_t count)
{
	if (count > srvtab_count)
		count = srvtab_count;
	for (; i < count; i++)
		format_srv(srvtab[i]);
	return count;
}

void
slb_fmtout()
{
	size_t n = 0;
	
	if (!srvtab_count) {
		logmsg(LOG_WARNING, "nothing to output");
		return;
	}

	slb_check_output();
	
	if (begin_output_message)
		fprintf(outfile, "%s", begin_output_message);
	if (output_head_count || output_tail_count) {
		if (output_head_count)
			n = format_output(0, output_head_count);
		if (output_tail_count) {
			size_t first = (srvtab_count < output_tail_count) ?
				       0 : (srvtab_count - output_tail_count);
			if (first < n)
				first = n;
			format_output(first, srvtab_count);
		}
	} else
		format_output(0, srvtab_count);
	if (end_output_message)
		fprintf(outfile, "%s", end_output_message);
	fflush(outfile);
}

#if defined HAVE_SYSCONF && defined _SC_OPEN_MAX
# define getmaxfd() sysconf(_SC_OPEN_MAX)
#elif defined (HAVE_GETDTABLESIZE)
# define getmaxfd() getdtablesize()
#else
# define getmaxfd() 256
#endif

void
close_fds(fd_set *fdset)
{
	int i;
	
	for (i = getmaxfd(); i >= 0; i--) {
		if (fdset && FD_ISSET(i, fdset))
			continue;
		close(i);
	}
}

int
open_redirector(const char *tag, int prio, int pidn)
{
	int p[2];
	FILE *fp;
	char buf[1024];
	pid_t pid;
	fd_set fdset;
	
	pipe(p);
	switch (pid = fork()) {
	case 0:
		/* Redirector process */
		FD_ZERO(&fdset);
		FD_SET(p[0], &fdset);
		if (log_to_stderr)
			FD_SET(2, &fdset);
		close_fds(&fdset);
		set_signals(SIG_DFL, slb_sigtab, slb_sigcount);
      
		close(p[1]);
		fp = fdopen(p[0], "r");
		if (fp == NULL)
			_exit(1);
		syslog_tag = tag;
		logger_setup();
		while (fgets(buf, sizeof(buf), fp)) {
			int cont;
			char *p = buf + strlen(buf) - 1;
			if (*p == '\n') {
				*p = 0;
				cont = 0;
			} else
				cont = 1;
			if (cont)
				logmsg(prio, "%s\\", buf);
			else
				logmsg(prio, "%s", buf);
		}
		_exit(0);
      
	case -1:
		logmsg(LOG_CRIT,
		       "cannot run redirector `%s': fork failed: %s",
		       tag, strerror(errno));
		return -1;

	default:
		debug(1, ("redirector for %s started, pid=%lu",
		      tag, (unsigned long) pid));
		proctab[pidn].pid = pid;
		close(p[0]);
		return p[1];
	}
}

static void
sigchld(int sig)
{
	pid_t pid;
	int status;
	
	if ((pid = waitpid((pid_t)-1, &status, WNOHANG)) != -1) {
		int i;
		for (i = 0; i < _PID_MAX; i++)
			if (proctab[i].pid == pid) {
				proctab[i].pid = -1;
				proctab[i].status = status;
			}
	}
}
	
static void
waitforproc(int pidn)
{
	int i;
	
	for (i = 0; i < 10; i++) {
		if (proctab[pidn].pid == -1) {
			int status = proctab[pidn].status;
			if (WIFEXITED(status))
				debug(1,
				      ("%s finished with status %d",
				       proctab[pidn].descr,
				       WEXITSTATUS(status)));
			else if (WIFSIGNALED(status))
				debug(1,
				      ("%s terminated on signal %d",
				       proctab[pidn].descr,
				       WTERMSIG(status)));
			else
				debug(1, ("%s terminated",
					  proctab[pidn].descr));
			return;
		}
		sleep(1);
	}
	logmsg(LOG_NOTICE, "%s did not finish; killing it",
	       proctab[pidn].descr);
	
	kill(proctab[pidn].pid, SIGKILL);
}

static void
open_pipe(const char *cmd)
{
	int err, out;
	int p[2];
	pid_t pid;
	fd_set fdset;
	int sigtab[] = { SIGCHLD };
	char *argv[4];
	
	while (*cmd && c_isspace(*cmd))
		cmd++;

	if (*cmd == 0) {
		logmsg(LOG_CRIT, "empty output command");
		exit(EX_CONFIG);
	}

	set_signals(sigchld, sigtab, sizeof(sigtab) / sizeof(sigtab[0]));
	
	pipe(p);
	out = open_redirector("stdout", LOG_INFO, PID_OUT);
	err = open_redirector("stderr", LOG_ERR, PID_ERR);

	switch (pid = fork()) {
		/* The child branch.  */
	case 0:
		set_signals(SIG_DFL, slb_sigtab, slb_sigcount);

		dup2(p[0], 0);
		dup2(out, 1);
		dup2(err, 2);

		FD_ZERO(&fdset);
		FD_SET(0, &fdset);
		FD_SET(1, &fdset);
		FD_SET(2, &fdset);
		close_fds(&fdset);

		argv[0] = "/bin/sh";
		argv[1] = "-c";
		argv[2] = (char*) cmd;
		argv[3] = NULL;
		execv(argv[0], argv);
		_exit(127);

	case -1:
		logmsg(LOG_CRIT,
		       "cannot run `%s': fork failed: %s",
			cmd, strerror(errno));
		exit(EX_UNAVAILABLE);

	default:
		proctab[PID_CMD].pid = pid;
		debug(1, ("%s started, pid=%lu", cmd, (unsigned long) pid));
	}
	outfile = fdopen(p[1], "w");
	close(p[0]);
}

static void
open_file(const char *file)
{
	outfile = fopen(file, "w");
	if (!outfile) {
		logmsg(LOG_CRIT, "cannot open output file %s: %s",
		       file, strerror(errno));
		exit(EX_UNAVAILABLE);
	}
}

static void
file_closefn()
{
	fclose(outfile);
}

static void
proc_closefn()
{
	fclose(outfile);
	waitforproc(PID_CMD);
	if (proctab[PID_ERR].pid != -1) {
		kill(proctab[PID_ERR].pid, SIGTERM);
		waitforproc(PID_ERR);
	}
	if (proctab[PID_OUT].pid != -1) {
		kill(proctab[PID_OUT].pid, SIGTERM);
		waitforproc(PID_OUT);
	}
}

static void
proc_checkfn()
{
	if (proctab[PID_CMD].pid == -1) {
		fclose(outfile);
		outfile = NULL;
		waitforproc(PID_CMD);
		if (proctab[PID_ERR].pid != -1) {
			kill(proctab[PID_ERR].pid, SIGTERM);
			waitforproc(PID_ERR);
		}
		if (proctab[PID_OUT].pid != -1) {
			kill(proctab[PID_OUT].pid, SIGTERM);
			waitforproc(PID_OUT);
		}

		open_pipe(output_file + 1);
	}
}

void (*closeoutfn)(void);
void (*checkoutfn)(void);

void
slb_open_output()
{
	if (strcmp(output_file, "-") == 0)
		outfile = stdout;
	else if (*output_file == '|') {
		open_pipe(output_file + 1);
		closeoutfn = proc_closefn;
		checkoutfn = proc_checkfn;
	} else {
		open_file(output_file);
		closeoutfn = file_closefn;
	}
}

void
slb_close_output()
{
	if (closeoutfn)
		closeoutfn();
}

void
slb_check_output()
{
	if (checkoutfn)
		checkoutfn();
}
